"""
OpShin provides some helper classes that define concepts introduced in PlutusTx
and used by the cardano node to encode data.
In particular you find the definition of the entire Script Context for Plutus V2
in the file `api_v2`.

```python
from opshin.ledger.api_v2 import *
```
"""
